/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2014, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */
#include <string.h>

#include "gstdwlallocator.h"

GST_DEBUG_CATEGORY_EXTERN (GST_CAT_PERFORMANCE);
GST_DEBUG_CATEGORY_EXTERN (GST_CAT_MEMORY);

GST_DEBUG_CATEGORY_STATIC (gst_dwl_allocator_debug);
#define GST_CAT_DEFAULT gst_dwl_allocator_debug

static GstAllocator *_dwl_allocator = NULL;

#define DWL_FAILED(ret) (DWL_OK != (ret))

static GstMemory *gst_dwl_allocator_alloc (GstAllocator * allocator, gsize size,
    GstAllocationParams * params);
static void gst_dwl_allocator_free (GstAllocator * allocator,
    GstMemory * memory);

#define GST_DWL_ALLOCATOR(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_DWL_ALLOCATOR,GstDwlAllocator))
#define GST_DWL_ALLOCATOR_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_DWL_ALLOCATOR,GstDwlAllocatorClass))

typedef struct _GstDwlMemory
{
  GstG1Memory mem;

  DWLLinearMem_t linearmem;
} GstDwlMemory;

typedef struct
{
  GstG1Allocator parent;

  gpointer dwl;
} GstDwlAllocator;

typedef struct
{
  GstG1AllocatorClass parent_class;
} GstDwlAllocatorClass;

G_DEFINE_TYPE (GstDwlAllocator, gst_dwl_allocator, GST_TYPE_G1_ALLOCATOR);

void
gst_dwl_allocator_new (void)
{
  if (_dwl_allocator) {
    GST_DEBUG ("allocator already registered");
    goto exit;
  }

  _dwl_allocator = g_object_new (gst_dwl_allocator_get_type (), NULL);
  if (!_dwl_allocator) {
    GST_ERROR ("unable to create dwl allocator");
    goto exit;
  }

  gst_allocator_register (GST_ALLOCATOR_DWL, _dwl_allocator);
  GST_DEBUG (GST_ALLOCATOR_DWL " successfully registered");

exit:
  {
    return;
  }
}

static void
gst_dwl_allocator_class_init (GstDwlAllocatorClass * klass)
{
  GstAllocatorClass *allocator_class;

  allocator_class = (GstAllocatorClass *) klass;

  allocator_class->alloc = GST_DEBUG_FUNCPTR (gst_dwl_allocator_alloc);
  allocator_class->free = GST_DEBUG_FUNCPTR (gst_dwl_allocator_free);

  GST_DEBUG_CATEGORY_INIT (gst_dwl_allocator_debug, "g1allocator",
      0, "G1 Memory Allocator");
}

static void
gst_dwl_allocator_init (GstDwlAllocator * allocator)
{
  DWLInitParam_t params;

  GST_CAT_DEBUG (GST_CAT_MEMORY, "init allocator %p", allocator);

  /* Use H264 as client, not really needed for anything but as a container */
  params.clientType = DWL_CLIENT_TYPE_H264_DEC;
  allocator->dwl = DWLInit (&params);
}

static GstMemory *
gst_dwl_allocator_alloc (GstAllocator * allocator, gsize size,
    GstAllocationParams * params)
{
  GstDwlAllocator *dwl;
  GstDwlMemory *dwlmem;
  GstG1Memory *g1mem;
  GstMemory *mem;
  gsize maxsize;
  gint ret;

  dwl = GST_DWL_ALLOCATOR (allocator);
  dwlmem = g_slice_new (GstDwlMemory);
  mem = GST_MEMORY_CAST (dwlmem);

  /* Take into account prefix and padding */
  maxsize = size + params->prefix + params->padding;

  GST_LOG ("Allocating new slice %p of %d", mem, maxsize);

  ret = DWLMallocLinear (dwl->dwl, maxsize, &dwlmem->linearmem);
  if (DWL_FAILED (ret)) {
    GST_ERROR_OBJECT (dwl, "Unable to allocate buffer of size %d, reason: %d",
        size, ret);
    g_free (dwlmem);
    dwlmem = NULL;
    mem = NULL;
    goto exit;
  }

  /* Initialize GstMemory */
  gst_memory_init (mem, GST_MEMORY_FLAG_PHYSICALLY_CONTIGUOUS, _dwl_allocator,
      NULL, maxsize, 0, params->prefix, size);

  g1mem = (GstG1Memory *) dwlmem;
  g1mem->virtaddress = dwlmem->linearmem.virtualAddress;
  g1mem->physaddress = dwlmem->linearmem.busAddress;

exit:
  {
    return mem;
  }
}

static void
gst_dwl_allocator_free (GstAllocator * allocator, GstMemory * mem)
{
  GstDwlAllocator *dwl;
  GstDwlMemory *dwlmem;

  dwl = GST_DWL_ALLOCATOR (allocator);
  dwlmem = (GstDwlMemory *) mem;

  g_return_if_fail (GST_IS_DWL_ALLOCATOR (mem->allocator));

  GST_LOG ("Freeing slice %p", mem);

  DWLFreeLinear (dwl->dwl, &dwlmem->linearmem);
  g_slice_free (GstDwlMemory, dwlmem);
}
